/*
 * Copyright (C) 2022  Aravinth Manivannan <realaravinth@batsense.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
use std::env;
use std::path::Path;

use config::{Config, ConfigError, Environment, File};
use derive_more::Display;
use log::info;
use log::warn;
use serde::Deserialize;
use serde::Serialize;
use url::Url;

const PREFIX: &str = "LPCONDUCTOR";
const SEPARATOR: &str = "_";

#[derive(Debug, Clone, Deserialize)]
pub struct Server {
    pub port: u32,
    pub domain: String,
    pub ip: String,
    pub url_prefix: Option<String>,
    pub proxy_has_tls: bool,
}

impl Server {
    #[cfg(not(tarpaulin_include))]
    pub fn get_ip(&self) -> String {
        format!("{}:{}", self.ip, self.port)
    }
}

#[derive(Deserialize, Serialize, Display, Eq, PartialEq, Clone, Debug)]
#[serde(rename_all = "lowercase")]
pub enum ConductorType {
    /// doesn't do anything
    #[display(fmt = "dummy")]
    Dummy,
}

#[derive(Debug, Clone, Deserialize)]
pub struct Creds {
    pub username: String,
    pub password: String,
}

#[derive(Debug, Clone, Deserialize)]
pub struct Settings {
    pub debug: bool,
    pub api_keys: Vec<Creds>,
    pub server: Server,
    pub source_code: String,
    pub conductor: ConductorType,
}

#[cfg(not(tarpaulin_include))]
impl Settings {
    pub fn authenticate(&self, username: &str, password: &str) -> bool {
        self.api_keys
            .iter()
            .any(|c| c.username == username && c.password == password)
    }

    pub fn new() -> Result<Self, ConfigError> {
        let mut s = Config::new();

        const CURRENT_DIR: &str = "./config/config.toml";
        const ETC: &str = "/etc/lpconductor/config.toml";

        if let Ok(path) = env::var("LPCONDUCTOR_CONFIG") {
            s.merge(File::with_name(&path))?;
        } else if Path::new(CURRENT_DIR).exists() {
            // merging default config from file
            s.merge(File::with_name(CURRENT_DIR))?;
        } else if Path::new(ETC).exists() {
            s.merge(File::with_name(ETC))?;
        } else {
            warn!("configuration file not found");
        }

        s.merge(Environment::with_prefix(PREFIX).separator(SEPARATOR))?;
        set_separator_field(&mut s);

        check_url(&s);

        match env::var("PORT") {
            Ok(val) => {
                s.set("server.port", val).unwrap();
            }
            Err(e) => warn!("couldn't interpret PORT: {}", e),
        }

        match s.try_into::<Self>() {
            Ok(val) => {
                Ok(val)
            },
            Err(e) => Err(ConfigError::Message(format!("\n\nError: {}. If it says missing fields, then please refer to https://git.batsense.net/LibrePages/conductor to learn more about how conductor reads configuration\n\n", e))),
        }
    }
}

#[cfg(not(tarpaulin_include))]
fn set_separator_field(s: &mut Config) {
    // ref: https://github.com/mehcode/config-rs/issues/391

    fn from_env(s: &mut Config, env_name: &str, config_name: &str) {
        if let Ok(val) = env::var(env_name) {
            info!("Overriding {config_name} with data from env var {env_name}");
            s.set(config_name, val)
                .unwrap_or_else(|_| panic!("Couldn't set {config_name}  from env var {env_name}"));
        }
    }
    from_env(s, &format!("{PREFIX}{SEPARATOR}SOURCE_CODE"), "source_code");
    from_env(
        s,
        &format!("{PREFIX}{SEPARATOR}SERVER{SEPARATOR}URL_PREFIX"),
        "server.url_prefix",
    );
    from_env(
        s,
        &format!("{PREFIX}{SEPARATOR}SERVER{SEPARATOR}PROXY_HAS_TLS"),
        "server.proxy_has_tls",
    );
}

#[cfg(not(tarpaulin_include))]
fn check_url(s: &Config) {
    let url = s
        .get::<String>("source_code")
        .expect("Couldn't access source_code");

    Url::parse(&url).expect("Please enter a URL for source_code in settings");
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn creds_works() {
        let settings = Settings::new().unwrap();
        let mut creds = settings.api_keys.get(0).unwrap().clone();

        assert!(settings.authenticate(&creds.username, &creds.password));

        creds.username = "noexist".into();
        assert!(!settings.authenticate(&creds.username, &creds.password));

        let mut creds = settings.api_keys.get(0).unwrap().clone();

        creds.password = "noexist".into();
        assert!(!settings.authenticate(&creds.username, &creds.password));
    }
}
